# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.

import jax

#: JAX Array
Array = jax.Array

#: JAX PRNGKey
PRNGKey = jax.Array
